"""A vector search engine in Python.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

This is a simple implementation of a vector-based search engine, based
on an article written by [add author's name]

I thought it might be fun to implement this in Python, so here it is.
I'll try to heavily document this code so that others can play around
with it.

There are a few external modules that I'm depending on; I use Roman
Geus's 'pysparse' sparse matrix multiplication library to represent
the huge sparse document matrix.  I'm also doing a bunch of matrix
stuff based on Numeric Python.
"""


from __future__ import generators
import spmatrix
import Numeric
import sys
import os
import shelve




class Engine:
    def __init__(self, documents,
                 tokenizer=None,
                 file_prefix="vector_search",
                 logfile=sys.stderr,
                 cache=True):
        if tokenizer == None:
            import tokenizer
            self.tokenizer = tokenizer.simpleTokenizer

        if os.path.isfile(file_prefix + ".engine"):
            self._restoreEngine(file_prefix)
        else:
            self.vectorizer = Vectorizer(self.tokenizer)
            self.logfile = logfile
            self.vectorizer.primeEngine(documents)
            self.doc_id_map = {}

        if os.path.isfile(file_prefix + ".matrix"):
            self._restoreMatrix(file_prefix)
        else:
            self.matrix = spmatrix.ll_mat(self.vectorizer.getWordCount(),
                                          len(documents))
            for (id, (doc_id, doc_text)) in enumerate(documents):
                self.log("processing %s" % doc_id)
                self.matrix[:, id] = self.vectorizer.makeSparseVector(doc_text)
                self.doc_id_map[id] = doc_id

        if cache:
            self._save(file_prefix)
        self.matrix = self.matrix.to_csr()


    def mapToDocId(self, id):
        return self.doc_id_map[id]


    def log(self, s):
        """Writes out to the current log file."""
        self.logfile.write(s)
        self.logfile.write("\n")
        
        
    def _save(self, file_prefix):
        """Saves our engine to disk in two separate files: our words
        dictionary, and our matrix."""
        self.matrix.export_mtx(file_prefix + ".matrix")
        d = shelve.open(file_prefix + ".engine")
        d['vectorizer'] = self.vectorizer
        d['doc_id_map'] = self.doc_id_map
        d.close()


    def _restoreEngine(self, file_prefix):
        d = shelve.open(file_prefix + ".engine")
        self.vectorizer = d['vectorizer']
        self.doc_id_map = d['doc_id_map']


    def _restoreMatrix(self, file_prefix):
        self.matrix = spmatrix.ll_mat_from_mtx(file_prefix + ".matrix")
        


    def search(self, query_text, threshold=0.5):
        query_vector = self.vectorizer.makeVector(query_text)
        result_vector = Numeric.zeros(self.matrix.shape[1],
                                      typecode='d')
        self.matrix.matvec_transp(query_vector, result_vector)

        return map(self.mapToDocId,
                   Numeric.nonzero(Numeric.greater(result_vector, threshold)))



class Vectorizer:
    """The vectorizer takes a document, and transforms it into a sparse
    vector."""
    def __init__(self, tokenize):
        self.tokenize = tokenize
        self.count_words = 0
        self.words_seen = {}

    def getWordCount(self): return self.count_words

    def primeEngine(self, documents):
        """Prepares our vectorizer by showing it our sample of
        documents.  This teaches it which words to pay attention
        to."""
        for id, text in documents:
            words = self.tokenize(text)
            for w in words:
                if w not in self.words_seen:
                    self.words_seen[w] = self.count_words
                    self.count_words += 1


    def makeVector(self, doc):
        """Given a document, returns a Numeric matrix that represents
        that document."""
        vector = Numeric.zeros(self.count_words)
        for word in self.tokenize(doc):
            id = self.words_seen.get(word, -1)
            if id != -1:
                vector[id] += 1
        return vector / (Numeric.dot(vector, vector)**(0.5))
    

    def makeSparseVector(self, doc):
        """Given a document, produces a sparse vector that represents
        that document."""
        vector = spmatrix.ll_mat(self.count_words, 1)
        for word in self.tokenize(doc):
            id = self.words_seen.get(word, -1)
            if id != -1:
                vector[id, 0] += 1
        return norm(vector)


######################################################################
"""Here's a small utility function; perhaps I'm not looking closely
enough, but I couldn't find a vector normalization function in
spmatrix."""

def norm(vec):
    """Produces the norm of a vector: the length of our vector becomes 1.

    Abnormal case: if the vector is the zero vector, we just return it
    back.
    
    """
    scaling_matrix = spmatrix.dot(vec, vec)
    if scaling_matrix[0,0] == 0:
        return vec
    else:
        scaling_matrix[0,0] = 1/(scaling_matrix[0,0]**(0.5))
        return spmatrix.matrixmultiply(vec, scaling_matrix)


def enumerate(seq):
    i = 0
    for x in seq:
        yield (i, x)
        i += 1
